#include "astropulse.h"
#ifdef BOINC_APP_GRAPHICS
#include "ap_graphics.h"
#endif
#include "fftw3.h"
#include "sbtf.h"
#include "ap_debug.h"
#include <stdio.h>
#include "mtrand.h"

namespace Astropulse {

void Science::print_DC_strengths() {

  const int NUM_STRENGTHS = 8;
  float strengths[NUM_STRENGTHS];
  float totals[NUM_STRENGTHS];
  float averages[NUM_STRENGTHS];
  float maxima[NUM_STRENGTHS];
  // long fft_len = client.ap_shmem->ap_gdata.state.fft_len;
  int fft_len2 = 1024;
  // int fft_len2 = 32768;
  long datasize = 8 * 1024 * 1024;
  FILE* outfile = fopen("temp.txt", "w");
  int num_ffts = 0;

  for(int i = 0; i < NUM_STRENGTHS; i++) {
    totals[i] = 0;
    maxima[i] = 0;
  }

  // Note rawdata is in BYTES, each byte is 4 samples, fft_len is in samples, so # bytes = fft_len2 / 4
  // data_chunk_now is also in bytes, but it INCREMENTS by fft_len2 /2 / 4 to produce overlap.
  // That is, it only increments by HALF of a FFT length each iteration.
  // To find # of samples from data_chunk_now, MULTIPLY BY 4!
  for(long data_chunk_now = 0; data_chunk_now < datasize - fft_len2/2/4; data_chunk_now += fft_len2/2/4) {
    num_ffts++;
    if (splitter_bits_to_float(&client.rawdata[data_chunk_now],
                             (float *)data, fft_len2)) {
      exit(SBTF_ERR);
    }
    fftwf_execute(fp2); // Note that this fft knows about fft_len2, and computes in chunks of fft_len2
			   // The maximum possible result is 1,024, average should be such that power = 1024 
    for(int i = 0; i < NUM_STRENGTHS; i++) { 
      strengths[i] = 0;
      long amt_to_subtract = (i==0)?0:pow_int(2,i-1);
      for(long j = 0; j < pow_int(2,i); j++) {
        float real = data[(j - amt_to_subtract + fft_len2) % fft_len2][0];
        float imag = data[(j - amt_to_subtract + fft_len2) % fft_len2][1];
        strengths[i] += real * real + imag * imag;
      }
      totals[i] += strengths[i];
      maxima[i] = (strengths[i] > maxima[i])?strengths[i]:maxima[i];
      // fprintf(outfile, "strengths[%d] = %f\n", i, strengths[i]);
      // fprintf(outfile, "totals[%d] = %f\n", i, totals[i]);
      // fprintf(outfile, "maxima[%d] = %f\n", i, maxima[i]);
      // fprintf(outfile, "strengths[%d] / %d = %f\n", i, pow_int(2, i), strengths[i] / pow(2, i));
      fflush(outfile);
    }
  }

  fprintf(outfile, "The averages below represent power per DC bin.\n");
  for(int i = 0; i < NUM_STRENGTHS; i++) {
    averages[i] = totals[i] / (float) num_ffts / (float) pow_int(2,i) / (float) fft_len2 / 2.0;
    fprintf(outfile, "averages[%d] = %f\n", i, averages[i]);
  } 
  for(int i = 0; i < NUM_STRENGTHS; i++) {
    fprintf(outfile, "normalized maxima[%d] = %f\n", i, maxima[i] / (float) pow_int(2,i) / (float) fft_len2 / 2.0);
  } 
  fflush(outfile);
}

int Science::get_indices_to_randomize(std::vector<long>& indices, int scale) {

  FILE* outfile;
  int init_indices = 100;
  int num_indices = 0;
  int size = init_indices;
  indices.resize(init_indices);
  if(scale > 7) {
      fprintf(stderr, "In Science::get_indices_to_randomize: error: scale > 7\n");
      printf("In Science::get_indices_to_randomize: error: scale > 7\n"); 
      exit(-1);
  }
  // power_threshold[scale]
  float power_threshold[] = { 12.1, 6.9, 5.3, 3.0, 2.46, 1.96, 1.62, 1.37};
  
  int fft_len2 = 1024;
  long datasize = 8 * 1024 * 1024;
  float strength;

  outfile = fopen("indices.txt", "w");

  for(long data_chunk_now = 0; data_chunk_now < datasize - fft_len2/2/4; data_chunk_now += fft_len2/2/4) {
    if (splitter_bits_to_float(&client.rawdata[data_chunk_now],
                             (float *)data, fft_len2)) {
      exit(SBTF_ERR);
    }
    fftwf_execute(fp2); // Note that this fft knows about fft_len2 (if you define it in astropulse.h), and computes in chunks of fft_len2
			   // The maximum possible result is 1,024, average should be such that power = 1024 
    
    /* For debugging: check power in DC bin for this data chunk
    if(data_chunk_now == 1359872) {
        printf("DC bin: %f %f, DC power: %f\n", data[0][0], data[0][1], (data[0][0] * data[0][0] + data[0][1] * data[0][1]) / fft_len2 / 2.0);
    }*/

    int expscale = pow_int(2, scale);
    strength = 0;
    for(int j = - expscale / 2; j < (expscale + 1) / 2; j++) {
      int j2 = (j + fft_len2) % fft_len2; 
      float real = data[j2][0];
      float imag = data[j2][1];
      strength += (real * real + imag * imag) / fft_len2 / 2.0 / expscale;
      // strength += (real * real + imag * imag) / fft_len2 / expscale;
    }

    /* For debugging: check that total power / 2.0 = fft_len2
    if(data_chunk_now < 10000) {
    float total = 0;
    for(int i = 0; i < fft_len2; i++) {
        total += (data[i][0] * data[i][0] + data[i][1] * data[i][1]) / fft_len2 / 2.0;    
    }
    printf("total = %f\n", total);
    }
    */

    if(strength > power_threshold[scale]) {
      fprintf(outfile, "At sample %d, data_chunk_now %d, strength %f > %f\n", data_chunk_now * 4, data_chunk_now, strength, power_threshold[scale]);
      indices[num_indices++] = data_chunk_now;
      if(num_indices >= size)
        size = num_indices + init_indices;
        indices.resize(size);
    }
  }

  fclose(outfile);
  return num_indices;
}


void Science::randomize_indices(const std::vector<long>& indices, int num_indices, int data_chunk_now, int fft_len, state_t& state) { 

// FILE* randsampfile = fopen("randsamp.txt", "a");
long prev_index = data_chunk_now - 1;
long prev_sample = (data_chunk_now - 1) * 4;
long this_index;
long this_sample;
   
    /* For debugging: print values of bits before they are randomized. 
    if(data_chunk_now == 1359872 - 4096) {
        printf("data_chunk_now: 1359872 - 4096\n");
        for(int k = 16384; k < 16384 + 1024; k++) {
            printf("%f %f | ", data[k][0], data[k][1]);
            if(k % 16 == 0) printf("\n");
        }
    }
    */
  
  // printf("data_chunk_now = %d\n", data_chunk_now); // for debugging
  for(int current = 0; current < num_indices; current++) {
    // If we're within 100000 bytes of a DC pulse, randomize this data chunk.
    if(indices[current] < data_chunk_now + 100000 && indices[current] > data_chunk_now - 100000) {
      // if we just blanked the last data chunk, then this data chunk only
      // adds fft_len / 2 samples, because of the overlap.
      if(state.lastDataChunkBlanked == data_chunk_now - fft_len / 4 / 2)
        state.numSamplesBlankedNew += fft_len / 2;
      else
        state.numSamplesBlankedNew += fft_len;
      double percentBlanked = (double) state.numSamplesBlankedNew / NUM_SAMPLES;
      state.lastDataChunkBlankedNew = data_chunk_now;
      long lastSampleBlanked = state.lastDataChunkBlankedNew * 4L + fft_len - 1;
      int debug = 0;
      randomize_samples(fft_len, state.mtr_new, debug);
      // fprintf(randsampfile, "Num samples blanked: %ld, percent blanked: %f, last sample blanked: %ld\n", state.numSamplesBlankedNew, 
      //       percentBlanked, lastSampleBlanked);
      break;
    }
  }
  // fclose(randsampfile);
}

// Use debug flag if you want randomize_samples to print something out
// on specific samples, for instance.
void Science::randomize_samples(int fft_len, MtRand& mtr, int debug) {
  // FILE *randsampfile;
  // randsampfile = fopen("randsamp.txt", "a");
  // fprintf(randsampfile, "in randomize_samples: in: %s\n", mtr.toString());

  for(int index = 0; index < fft_len; index++) {
    data[index][0] = mtr.drand32() > 0.5 ? 1.0 : - 1.0; 
    data[index][1] = mtr.drand32() > 0.5 ? 1.0 : - 1.0; 
  }

  // fprintf(randsampfile, "in randomize_samples: out: %s\n", mtr.toString());
  // fflush(randsampfile);
  // fclose(randsampfile);
}

} // end namespace Astropulse
